@@ -132,6 +132,7 @@ editor_save_changes_dialog_discard (GtkMessageDialog *dialog,
       const SaveRequest *sr = &g_array_index (requests, SaveRequest , i);
 
       _editor_page_discard_changes_async (sr->page,
+                                          FALSE,
                                           NULL,
                                           editor_save_changes_dialog_discard_cb,
                                           g_array_ref (requests));
@@ -171,6 +172,7 @@ editor_save_changes_dialog_save_cb (GObject      *object,
             editor_save_changes_dialog_remove (requests, i);
           else
             _editor_page_discard_changes_async (sr->page,
+                                                FALSE,
                                                 NULL,
                                                 editor_save_changes_dialog_discard_cb,
                                                 g_array_ref (requests));
@@ -243,6 +245,7 @@ _editor_save_changes_dialog_new (GtkWindow *parent,
   GtkWidget *dialog;
   GtkWidget *group;
   GtkWidget *area;
+  GtkWidget *button;
 
   g_return_val_if_fail (!parent || GTK_IS_WINDOW (parent), NULL);
   g_return_val_if_fail (pages != NULL, NULL);
@@ -252,7 +255,7 @@ _editor_save_changes_dialog_new (GtkWindow *parent,
   requests = g_array_new (FALSE, FALSE, sizeof (SaveRequest));
   g_array_set_clear_func (requests, save_request_clear);
 
-  discard_label = g_dngettext (GETTEXT_PACKAGE, _("Discard"), _("Discard All"), pages->len);
+  discard_label = g_dngettext (GETTEXT_PACKAGE, _("_Discard"), _("_Discard All"), pages->len);
 
   dialog = gtk_message_dialog_new (parent,
                                    GTK_DIALOG_MODAL | GTK_DIALOG_USE_HEADER_BAR,
@@ -267,6 +270,12 @@ _editor_save_changes_dialog_new (GtkWindow *parent,
                           _("_Save"), GTK_RESPONSE_YES,
                           NULL);
 
+  button = gtk_dialog_get_widget_for_response (GTK_DIALOG (dialog), GTK_RESPONSE_NO);
+  gtk_widget_add_css_class (button, "destructive-action");
+
+  button = gtk_dialog_get_widget_for_response (GTK_DIALOG (dialog), GTK_RESPONSE_YES);
+  gtk_widget_add_css_class (button, "suggested-action");
+
   area = gtk_message_dialog_get_message_area (GTK_MESSAGE_DIALOG (dialog));
   group = adw_preferences_group_new ();
   gtk_box_append (GTK_BOX (area), group);
@@ -282,11 +291,8 @@ _editor_save_changes_dialog_new (GtkWindow *parent,
       g_autofree gchar *subtitle_str = NULL;
       GFile *file = editor_document_get_file (document);
       g_autoptr(GFile) directory = file ? g_file_get_parent (file) : NULL;
-      GtkWidget *box;
       GtkWidget *check;
       GtkWidget *row;
-      GtkWidget *subtitle;
-      GtkWidget *title;
       SaveRequest sr;
 
       if (file == NULL)
@@ -314,29 +320,14 @@ _editor_save_changes_dialog_new (GtkWindow *parent,
         subtitle_str = _editor_path_collapse (g_file_peek_path (directory));
 
       row = adw_action_row_new ();
-      box = g_object_new (GTK_TYPE_BOX,
-                          "orientation", GTK_ORIENTATION_VERTICAL,
-                          "spacing", 3,
-                          "valign", GTK_ALIGN_CENTER,
-                          NULL);
-      title = g_object_new (GTK_TYPE_LABEL,
-                            "halign", GTK_ALIGN_START,
-                            "label", title_str,
-                            NULL);
-      subtitle = g_object_new (GTK_TYPE_LABEL,
-                               "halign", GTK_ALIGN_START,
-                               "label", subtitle_str,
-                               "attributes", smaller,
-                               NULL);
       check = g_object_new (GTK_TYPE_CHECK_BUTTON,
                             "active", TRUE,
                             NULL);
 
       gtk_accessible_update_property (GTK_ACCESSIBLE (check), GTK_ACCESSIBLE_PROPERTY_LABEL, _("Save changes for this document"), -1);
-      gtk_box_append (GTK_BOX (box), title);
-      gtk_box_append (GTK_BOX (box), subtitle);
+      adw_preferences_row_set_title (ADW_PREFERENCES_ROW (row), title_str);
+      adw_action_row_set_subtitle (ADW_ACTION_ROW (row), subtitle_str);
       adw_action_row_add_prefix (ADW_ACTION_ROW (row), check);
-      adw_action_row_add_suffix (ADW_ACTION_ROW (row), box);
       adw_action_row_set_activatable_widget (ADW_ACTION_ROW (row), check);
       adw_preferences_group_add (ADW_PREFERENCES_GROUP (group), row);
 
