@@ -28,10 +28,13 @@
 #include "editor-document-private.h"
 #include "editor-page-private.h"
 #include "editor-session-private.h"
+#include "editor-sidebar-item-private.h"
 #include "editor-window-private.h"
 
 #define DEFAULT_AUTO_SAVE_TIMEOUT_SECONDS 3
 #define MAX_AUTO_SAVE_TIMEOUT_SECONDS (60*5)
+#define MAX_BOOKMARKS 100
+#define HUGE_WINDOW 8000
 
 typedef struct
 {
@@ -55,6 +58,12 @@ typedef struct
   Position end;
 } Selection;
 
+typedef struct
+{
+  char *uri;
+  GDateTime *age;
+} Recent;
+
 G_DEFINE_TYPE (EditorSession, editor_session, G_TYPE_OBJECT)
 
 enum {
@@ -62,6 +71,7 @@ enum {
   PROP_AUTO_SAVE,
   PROP_AUTO_SAVE_DELAY,
   PROP_RECENTS,
+  PROP_CAN_CLEAR_HISTORY,
   N_PROPS
 };
 
@@ -179,6 +189,16 @@ add_window_state (EditorSession   *self,
   g_assert (EDITOR_IS_SESSION (self));
   g_assert (builder != NULL);
 
+  if (default_width > 0 && default_height > 0)
+    {
+      g_variant_builder_open (builder, G_VARIANT_TYPE ("{sv}"));
+      g_variant_builder_add (builder, "s", "default-window-size");
+      g_variant_builder_open (builder, G_VARIANT_TYPE ("v"));
+      g_variant_builder_add (builder, "(uu)", default_width, default_height);
+      g_variant_builder_close (builder);
+      g_variant_builder_close (builder);
+    }
+
   g_variant_builder_open (builder, G_VARIANT_TYPE ("{sv}"));
   g_variant_builder_add (builder, "s", "windows");
   g_variant_builder_open (builder, G_VARIANT_TYPE ("v"));
@@ -198,8 +218,7 @@ add_window_state (EditorSession   *self,
         g_variant_builder_add_parsed (builder, "{'is-active', <%b>}", is_active);
 
       /* Store the window size */
-      width = gtk_widget_get_width (GTK_WIDGET (window));
-      height = gtk_widget_get_height (GTK_WIDGET (window));
+      gtk_window_get_default_size (GTK_WINDOW (window), &width, &height);
       g_variant_builder_add_parsed (builder,
                                     "{'size', <(%u,%u)>}",
                                     CLAMP (width, 0, 10000),
@@ -318,6 +337,14 @@ find_page_for_file (EditorSession *self,
   return NULL;
 }
 
+static gboolean
+_editor_session_get_can_clear_history (EditorSession *self)
+{
+  g_assert (EDITOR_IS_SESSION (self));
+
+  return self->can_clear_history;
+}
+
 static void
 editor_session_dispose (GObject *object)
 {
@@ -378,6 +405,10 @@ editor_session_get_property (GObject    *object,
       g_value_set_uint (value, editor_session_get_auto_save_delay (self));
       break;
 
+    case PROP_CAN_CLEAR_HISTORY:
+      g_value_set_boolean (value, _editor_session_get_can_clear_history (self));
+      break;
+
     case PROP_RECENTS:
       g_value_set_object (value, editor_session_get_recents (self));
       break;
@@ -451,6 +482,18 @@ editor_session_class_init (EditorSessionClass *klass)
                        1, MAX_AUTO_SAVE_TIMEOUT_SECONDS, DEFAULT_AUTO_SAVE_TIMEOUT_SECONDS,
                        (G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
 
+  /**
+   * EditorSession:can-clear-history:
+   *
+   * If the history can be cleared.
+   */
+  properties [PROP_CAN_CLEAR_HISTORY] =
+    g_param_spec_boolean ("can-clear-history",
+                          "Can Clear History",
+                          "If the history can be cleared",
+                          FALSE,
+                          (G_PARAM_READABLE | G_PARAM_STATIC_STRINGS));
+
   /**
    * EditorSession:recents:
    *
@@ -572,26 +615,27 @@ _editor_session_new (void)
 
 static gboolean
 get_default_size (EditorSession *self,
-                  guint         *width,
-                  guint         *height)
+                  int           *width,
+                  int           *height)
 {
-  GApplication *app;
-  GtkWindow *active;
+  const GList *windows;
 
   g_assert (EDITOR_IS_SESSION (self));
 
-  app = g_application_get_default ();
-  active = gtk_application_get_active_window (GTK_APPLICATION (app));
+  windows = gtk_application_get_windows (GTK_APPLICATION (EDITOR_APPLICATION_DEFAULT));
 
-  if (active)
+  *width = default_width;
+  *height = default_height;
+
+  for (const GList *iter = windows; iter; iter = iter->next)
     {
-      *width = gtk_widget_get_width (GTK_WIDGET (active));
-      *height = gtk_widget_get_height (GTK_WIDGET (active));
-    }
-  else
-    {
-      *width = default_width;
-      *height = default_height;
+      GtkWindow *window = iter->data;
+
+      if (EDITOR_IS_WINDOW (window))
+        {
+          gtk_window_get_default_size (window, width, height);
+          break;
+        }
     }
 
   return *width > 0 && *height > 0;
@@ -622,8 +666,8 @@ _editor_session_create_window_no_draft (EditorSession *self)
 {
   EditorWindow *window;
   gboolean resize;
-  guint width;
-  guint height;
+  int width;
+  int height;
 
   g_return_val_if_fail (EDITOR_IS_SESSION (self), NULL);
 
@@ -653,8 +697,8 @@ editor_session_create_window (EditorSession *self)
   g_autoptr(EditorDocument) document = NULL;
   EditorWindow *window;
   gboolean resize;
-  guint width;
-  guint height;
+  int width;
+  int height;
 
   g_return_val_if_fail (EDITOR_IS_SESSION (self), NULL);
 
@@ -674,6 +718,16 @@ editor_session_create_window (EditorSession *self)
   return window;
 }
 
+static void
+editor_session_document_changed_cb (EditorSession  *self,
+                                    EditorDocument *document)
+{
+  g_assert (EDITOR_IS_SESSION (self));
+  g_assert (EDITOR_IS_DOCUMENT (document));
+
+  _editor_session_mark_dirty (self);
+}
+
 /**
  * editor_session_add_page:
  * @self: an #EditorSession
@@ -693,6 +747,8 @@ editor_session_add_page (EditorSession *self,
                          EditorWindow  *window,
                          EditorPage    *page)
 {
+  EditorDocument *document;
+
   g_return_if_fail (EDITOR_IS_SESSION (self));
   g_return_if_fail (EDITOR_IS_WINDOW (window));
   g_return_if_fail (EDITOR_IS_PAGE (page));
@@ -703,6 +759,14 @@ editor_session_add_page (EditorSession *self,
   g_assert (window != NULL);
   g_assert (EDITOR_IS_WINDOW (window));
 
+  /* Update session state when the document changes */
+  document = editor_page_get_document (page);
+  g_signal_connect_object (document,
+                           "changed",
+                           G_CALLBACK (editor_session_document_changed_cb),
+                           self,
+                           G_CONNECT_SWAPPED);
+
   g_ptr_array_add (self->pages, g_object_ref (page));
 
   _editor_window_add_page (window, page);
@@ -717,16 +781,6 @@ editor_session_add_page (EditorSession *self,
   _editor_session_mark_dirty (self);
 }
 
-static void
-editor_session_document_changed_cb (EditorSession  *self,
-                                    EditorDocument *document)
-{
-  g_assert (EDITOR_IS_SESSION (self));
-  g_assert (EDITOR_IS_DOCUMENT (document));
-
-  _editor_session_mark_dirty (self);
-}
-
 /**
  * editor_session_add_document:
  * @self: an #EditorSession
@@ -758,12 +812,6 @@ editor_session_add_document (EditorSession  *self,
   page = editor_page_new_for_document (document);
   editor_session_add_page (self, window, page);
 
-  g_signal_connect_object (document,
-                           "changed",
-                           G_CALLBACK (editor_session_document_changed_cb),
-                           self,
-                           G_CONNECT_SWAPPED);
-
   return page;
 }
 
@@ -966,6 +1014,16 @@ _editor_session_remove_window (EditorSession *self,
   if (self->windows->len == 1 &&
       EDITOR_WINDOW (g_ptr_array_index (self->windows, 0)) == window)
     {
+      int width, height;
+
+      /* Save the width/height as default for future sessions */
+      gtk_window_get_default_size (GTK_WINDOW (window), &width, &height);
+      if (width > 0 && width < HUGE_WINDOW && height > 0 && height < HUGE_WINDOW)
+        {
+          default_width = width;
+          default_height = height;
+        }
+
       editor_session_save_async (self,
                                  NULL,
                                  editor_session_save_for_shutdown_cb,
@@ -997,6 +1055,30 @@ _editor_session_remove_window (EditorSession *self,
   _editor_session_mark_dirty (self);
 }
 
+static int
+recent_compare (gconstpointer a,
+                gconstpointer b)
+{
+  const Recent *ra = a;
+  const Recent *rb = b;
+  int ret;
+
+  if (ra->age == NULL && rb->age == NULL)
+    return strcmp (ra->uri, rb->uri);
+
+  if (ra->age == NULL)
+    return 1;
+
+  if (rb->age == NULL)
+    return -1;
+
+  ret = g_date_time_compare (ra->age, rb->age);
+
+  if (ret < 0) return 1;
+  else if (ret > 0) return -1;
+  else return 0;
+}
+
 static void
 editor_session_update_recent_worker (GTask        *task,
                                      gpointer      source_object,
@@ -1005,6 +1087,9 @@ editor_session_update_recent_worker (GTask        *task,
 {
   EditorSessionSave *save = task_data;
   g_autoptr(GSettings) settings = NULL;
+  g_autoptr(GBookmarkFile) bookmarks = NULL;
+  g_autofree gchar *filename = NULL;
+  g_autoptr(GError) error = NULL;
 
   g_assert (G_IS_TASK (task));
   g_assert (EDITOR_IS_SESSION (source_object));
@@ -1021,46 +1106,74 @@ editor_session_update_recent_worker (GTask        *task,
       return;
     }
 
-  if (save->seen != NULL || save->forgot != NULL)
+  bookmarks = g_bookmark_file_new ();
+  filename = get_bookmarks_filename ();
+
+  if (!g_bookmark_file_load_from_file (bookmarks, filename, &error))
     {
-      g_autoptr(GBookmarkFile) bookmarks = g_bookmark_file_new ();
-      g_autofree gchar *filename = get_bookmarks_filename ();
-      g_autoptr(GError) error = NULL;
+      if (!g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
+        g_warning ("Failed to load bookmarks file: %s", error->message);
+      g_clear_error (&error);
+    }
 
-      if (!g_bookmark_file_load_from_file (bookmarks, filename, &error))
+  /* Truncate to recent items to avoid gigantic lists */
+  if (g_bookmark_file_get_size (bookmarks) > MAX_BOOKMARKS)
+    {
+      g_autoptr(GArray) ar = g_array_new (FALSE, FALSE, sizeof (Recent));
+      gchar **uris;
+      gsize length;
+
+      uris = g_bookmark_file_get_uris (bookmarks, &length);
+
+      for (gsize i = 0; i < length; i++)
         {
-          if (!g_error_matches (error, G_FILE_ERROR, G_FILE_ERROR_NOENT))
-            g_warning ("Failed to load bookmarks file: %s", error->message);
-          g_clear_error (&error);
+          Recent r;
+
+          r.age = g_bookmark_file_get_visited_date_time (bookmarks, uris[i], NULL);
+          r.uri = g_steal_pointer (&uris[i]);
+
+          g_array_append_val (ar, r);
         }
 
-      if (save->seen != NULL)
-        {
-          for (guint i = 0; i < save->seen->len; i++)
-            {
-              GFile *file = g_ptr_array_index (save->seen, i);
-              g_autofree gchar *uri = g_file_get_uri (file);
-              g_bookmark_file_add_application (bookmarks, uri, NULL, NULL);
-            }
-        }
+      g_clear_pointer (&uris, g_free);
 
-      if (save->forgot != NULL)
-        {
-          for (guint i = 0; i < save->forgot->len; i++)
-            {
-              GFile *file = g_ptr_array_index (save->forgot, i);
-              g_autofree gchar *uri = g_file_get_uri (file);
-              g_bookmark_file_remove_item (bookmarks, uri, NULL);
-            }
-        }
+      g_array_sort (ar, recent_compare);
 
-      if (!g_bookmark_file_to_file (bookmarks, filename, &error))
+      for (gsize i = MAX_BOOKMARKS; i < ar->len; i++)
         {
-          g_task_return_error (task, g_steal_pointer (&error));
-          return;
+          const Recent *r = &g_array_index (ar, Recent, i);
+
+          g_debug ("Removing %s from recents", r->uri);
+          g_bookmark_file_remove_item (bookmarks, r->uri, NULL);
         }
     }
 
+  if (save->seen != NULL)
+    {
+      for (guint i = 0; i < save->seen->len; i++)
+        {
+          GFile *file = g_ptr_array_index (save->seen, i);
+          g_autofree gchar *uri = g_file_get_uri (file);
+          g_bookmark_file_add_application (bookmarks, uri, NULL, NULL);
+        }
+    }
+
+  if (save->forgot != NULL)
+    {
+      for (guint i = 0; i < save->forgot->len; i++)
+        {
+          GFile *file = g_ptr_array_index (save->forgot, i);
+          g_autofree gchar *uri = g_file_get_uri (file);
+          g_bookmark_file_remove_item (bookmarks, uri, NULL);
+        }
+    }
+
+  if (!g_bookmark_file_to_file (bookmarks, filename, &error))
+    {
+      g_task_return_error (task, g_steal_pointer (&error));
+      return;
+    }
+
   g_task_return_boolean (task, TRUE);
 }
 
@@ -1305,12 +1418,18 @@ editor_session_open (EditorSession           *self,
 void
 editor_session_open_files (EditorSession  *self,
                            GFile         **files,
-                           gint            n_files)
+                           gint            n_files,
+                           const char     *hint)
 {
+  EditorWindow *window = NULL;
+
   g_return_if_fail (EDITOR_IS_SESSION (self));
 
+  if (g_strcmp0 (hint, "new-window") == 0)
+    window = editor_session_create_window (self);
+
   for (guint i = 0; i < n_files; i++)
-    editor_session_open (self, NULL, files[i], NULL);
+    editor_session_open (self, window, files[i], NULL);
 }
 
 /**
@@ -1610,6 +1729,7 @@ editor_session_restore_v1 (EditorSession *self,
   GVariantIter iter;
   GVariant *window;
   gboolean had_failure = FALSE;
+  guint width, height;
 
   g_assert (EDITOR_IS_SESSION (self));
   g_assert (state != NULL);
@@ -1618,6 +1738,13 @@ editor_session_restore_v1 (EditorSession *self,
   if ((drafts = g_variant_lookup_value (state, "drafts", G_VARIANT_TYPE ("aa{sv}"))))
     editor_session_restore_v1_drafts (self, drafts);
 
+  if (g_variant_lookup (state, "default-window-size", "(uu)", &width, &height) &&
+      width < HUGE_WINDOW && height < HUGE_WINDOW)
+    {
+      default_width = width;
+      default_height = height;
+    }
+
   if (!self->restore_pages ||
       !(windows = g_variant_lookup_value (state, "windows", G_VARIANT_TYPE ("aa{sv}"))) ||
       g_variant_n_children (windows) == 0)
@@ -1627,10 +1754,9 @@ editor_session_restore_v1 (EditorSession *self,
   while (g_variant_iter_loop (&iter, "@a{sv}", &window))
     {
       g_autoptr(GVariant) pages = NULL;
-      guint width, height;
 
       if (!g_variant_lookup (window, "size", "(uu)", &width, &height) ||
-          width > 10000 || height > 10000)
+          width > HUGE_WINDOW || height > HUGE_WINDOW)
         {
           width = 0;
           height = 0;
@@ -1687,6 +1813,38 @@ failure:
   editor_session_create_window (self);
 }
 
+static void
+editor_session_recents_items_changed_cb (EditorSession      *self,
+                                         guint               position,
+                                         guint               removed,
+                                         guint               added,
+                                         EditorSidebarModel *model)
+{
+  gboolean can_clear_history = FALSE;
+  guint n_items;
+
+  g_assert (EDITOR_IS_SESSION (self));
+
+  n_items = g_list_model_get_n_items (G_LIST_MODEL (model));
+
+  for (guint i = 0; i < n_items; i++)
+    {
+      g_autoptr(EditorSidebarItem) item = g_list_model_get_item (G_LIST_MODEL (model), i);
+
+      if (!_editor_sidebar_item_get_is_modified (item))
+        {
+          can_clear_history = TRUE;
+          break;
+        }
+    }
+
+  if (can_clear_history != self->can_clear_history)
+    {
+      self->can_clear_history = can_clear_history;
+      g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_CAN_CLEAR_HISTORY]);
+    }
+}
+
 static void
 editor_session_restore (EditorSession *self,
                         GVariant      *state)
@@ -1699,6 +1857,12 @@ editor_session_restore (EditorSession *self,
 
   self->recents = _editor_sidebar_model_new (self);
 
+  g_signal_connect_object (self->recents,
+                           "items-changed",
+                           G_CALLBACK (editor_session_recents_items_changed_cb),
+                           self,
+                           G_CONNECT_SWAPPED);
+
   if (g_variant_lookup (state, "version", "u", &version) && version == 1)
     editor_session_restore_v1 (self, state);
   else
@@ -1833,10 +1997,12 @@ editor_session_load_recent_worker (GTask        *task,
   files = g_ptr_array_new_with_free_func (g_object_unref);
   uris = g_bookmark_file_get_uris (bookmarks, &len);
 
+
   for (gsize i = 0; i < len; i++)
     {
       const gchar *uri = uris[i];
       g_autoptr(GFile) file = g_file_new_for_uri (uri);
+      GDateTime *age = g_bookmark_file_get_visited_date_time (bookmarks, uri, NULL);
 
       if (g_file_is_native (file))
         {
@@ -1848,6 +2014,15 @@ editor_session_load_recent_worker (GTask        *task,
             }
         }
 
+      /* Track the age on the GFile without having to create a
+       * new intermediate structure.
+       */
+      if (age != NULL)
+        g_object_set_data_full (G_OBJECT (file),
+                                "AGE",
+                                g_date_time_ref (age),
+                                (GDestroyNotify)g_date_time_unref);
+
       g_ptr_array_add (files, g_steal_pointer (&file));
     }
 
@@ -2007,23 +2182,24 @@ _editor_session_remove_draft (EditorSession *self,
 
 void
 _editor_session_move_page_to_window (EditorSession *self,
-                                     EditorPage    *page,
+                                     EditorPage    *epage,
                                      EditorWindow  *window)
 {
   EditorWindow *old_window;
 
   g_return_if_fail (EDITOR_IS_SESSION (self));
-  g_return_if_fail (EDITOR_IS_PAGE (page));
+  g_return_if_fail (EDITOR_IS_PAGE (epage));
   g_return_if_fail (EDITOR_IS_WINDOW (window));
 
-  old_window = _editor_page_get_window (page);
+  old_window = _editor_page_get_window (epage);
 
   if (window != old_window)
     {
-      g_object_ref (page);
-      _editor_window_remove_page (old_window, page);
-      _editor_window_add_page (window, page);
-      g_object_unref (page);
+      AdwTabPage *page = adw_tab_view_get_page (old_window->tab_view, GTK_WIDGET (epage));
+
+      _editor_page_begin_move (epage);
+      adw_tab_view_transfer_page (old_window->tab_view, page, window->tab_view, 0);
+      _editor_page_end_move (epage);
     }
 }
 
@@ -2220,10 +2396,51 @@ _editor_session_set_restore_pages (EditorSession *self,
 
   if (self->did_restore)
     {
-      g_warning ("Calling %s() after restoring hsa no effect. Ignoring request.",
+      g_warning ("Calling %s() after restoring has no effect. Ignoring request.",
                  G_STRFUNC);
       return;
     }
 
   self->restore_pages = !!restore_pages;
 }
+
+/**
+ * _editor_session_clear_history:
+ * @self: an #EditorSession
+ *
+ * This clears the history (recently-used.xbel) and clears the
+ * panel of those items but will not clear/remove unsaved drafts
+ * as we don't want to lose that data, only history.
+ */
+void
+_editor_session_clear_history (EditorSession *self)
+{
+  g_autofree char *filename = NULL;
+  guint n_items;
+
+  g_return_if_fail (EDITOR_IS_SESSION (self));
+
+  filename = get_bookmarks_filename ();
+  g_unlink (filename);
+
+  n_items = g_list_model_get_n_items (G_LIST_MODEL (self->recents));
+
+  for (guint i = n_items; i > 0; i--)
+    {
+      g_autoptr(EditorSidebarItem) item = g_list_model_get_item (G_LIST_MODEL (self->recents), i-1);
+      const char *draft_id;
+      GFile *file;
+
+      /* Ignore modified files to avoid losing data */
+      if (_editor_sidebar_item_get_is_modified (item))
+        continue;
+
+      file = _editor_sidebar_item_get_file (item);
+      draft_id = _editor_sidebar_item_get_draft_id (item);
+
+      if (file != NULL)
+        _editor_sidebar_model_remove_file (self->recents, file);
+      else
+        _editor_sidebar_model_remove_draft (self->recents, draft_id);
+    }
+}
