@@ -32,6 +32,12 @@
 #define RUN_CHECKED        GSIZE_TO_POINTER(1)
 #define UPDATE_DELAY_MSECS 100
 #define UPDATE_QUANTA_USEC (G_USEC_PER_SEC/1000L*2) /* 2 msec */
+/* Keyboard repeat is 30 msec by default (see
+ * org.gnome.desktop.peripherals.keyboard repeat-interval) so
+ * we want something longer than that so we are likely
+ * to get removed/re-added on each repeat movement.
+ */
+#define INVALIDATE_DELAY_MSECS 100
 
 typedef struct
 {
@@ -56,6 +62,8 @@ struct _EditorTextBufferSpellAdapter
   GtkTextTag         *no_spell_check_tag;
 
   guint               cursor_position;
+  guint               incoming_cursor_position;
+  guint               queued_cursor_moved;
 
   gsize               update_source;
 
@@ -161,15 +169,6 @@ editor_text_buffer_spell_adapter_new (GtkTextBuffer      *buffer,
                        NULL);
 }
 
-static inline gboolean
-contains_iter (const GtkTextIter *begin,
-               const GtkTextIter *end,
-               const GtkTextIter *iter)
-{
-  return gtk_text_iter_compare (begin, iter) <= 0 &&
-         gtk_text_iter_compare (iter, end) <= 0;
-}
-
 static gboolean
 get_unchecked_start_cb (gsize                   offset,
                         const CjhTextRegionRun *run,
@@ -299,16 +298,26 @@ editor_text_buffer_spell_adapter_invalidate_all (EditorTextBufferSpellAdapter *s
 
   g_assert (EDITOR_IS_TEXT_BUFFER_SPELL_ADAPTER (self));
 
-  length = _cjh_text_region_get_length (self->region);
+  if (!self->enabled)
+    return;
 
-  if (length > 0)
+  /* We remove using the known length from the region */
+  if ((length = _cjh_text_region_get_length (self->region)) > 0)
     {
-      _cjh_text_region_replace (self->region, 0, length - 1, RUN_UNCHECKED);
+      _cjh_text_region_remove (self->region, 0, length - 1);
       editor_text_buffer_spell_adapter_queue_update (self);
     }
 
+  /* We add using the length from the buffer because if we were not
+   * enabled previously, the textregion would be empty.
+   */
   gtk_text_buffer_get_bounds (self->buffer, &begin, &end);
-  gtk_text_buffer_remove_tag (self->buffer, self->tag, &begin, &end);
+  if (!gtk_text_iter_equal (&begin, &end))
+    {
+      length = gtk_text_iter_get_offset (&end) - gtk_text_iter_get_offset (&begin);
+      _cjh_text_region_insert (self->region, 0, length, RUN_UNCHECKED);
+      gtk_text_buffer_remove_tag (self->buffer, self->tag, &begin, &end);
+    }
 }
 
 static void
@@ -357,6 +366,9 @@ invalidate_tag_region_cb (EditorTextBufferSpellAdapter *self,
   g_assert (GTK_IS_TEXT_TAG (tag));
   g_assert (GTK_IS_TEXT_BUFFER (buffer));
 
+  if (!self->enabled)
+    return;
+
   if (tag == self->no_spell_check_tag)
     {
       gsize begin_offset = gtk_text_iter_get_offset (begin);
@@ -367,6 +379,35 @@ invalidate_tag_region_cb (EditorTextBufferSpellAdapter *self,
     }
 }
 
+static void
+apply_error_style_cb (GtkSourceBuffer *buffer,
+                      GParamSpec      *pspec,
+                      GtkTextTag      *tag)
+{
+  GtkSourceStyleScheme *scheme;
+  GtkSourceStyle *style;
+
+  g_assert (GTK_SOURCE_IS_BUFFER (buffer));
+  g_assert (GTK_IS_TEXT_TAG (tag));
+
+  g_object_set (tag,
+                "underline", PANGO_UNDERLINE_ERROR,
+                "background-set", FALSE,
+                "foreground-set", FALSE,
+                "weight-set", FALSE,
+                "variant-set", FALSE,
+                "style-set", FALSE,
+                "indent-set", FALSE,
+                "size-set", FALSE,
+                NULL);
+
+  if ((scheme = gtk_source_buffer_get_style_scheme (buffer)))
+    {
+      if ((style = gtk_source_style_scheme_get_style (scheme, "def:misspelled-word")))
+        gtk_source_style_apply (style, tag);
+    }
+}
+
 static void
 editor_text_buffer_spell_adapter_set_buffer (EditorTextBufferSpellAdapter *self,
                                              GtkTextBuffer                *buffer)
@@ -392,6 +433,13 @@ editor_text_buffer_spell_adapter_set_buffer (EditorTextBufferSpellAdapter *self,
                                               "underline", PANGO_UNDERLINE_ERROR,
                                               NULL);
 
+      g_signal_connect_object (buffer,
+                               "notify::style-scheme",
+                               G_CALLBACK (apply_error_style_cb),
+                               self->tag,
+                               0);
+      apply_error_style_cb (GTK_SOURCE_BUFFER (buffer), NULL, self->tag);
+
       /* Track tag changes from the tag table and extract "no-spell-check"
        * tag from GtkSourceView so that we can avoid words with that tag.
        */
@@ -443,6 +491,8 @@ editor_text_buffer_spell_adapter_set_enabled (EditorTextBufferSpellAdapter *self
         }
 
       editor_text_buffer_spell_adapter_invalidate_all (self);
+      editor_text_buffer_spell_adapter_queue_update (self);
+
       g_object_notify_by_pspec (G_OBJECT (self), properties [PROP_ENABLED]);
     }
 }
@@ -657,7 +707,8 @@ editor_text_buffer_spell_adapter_before_insert_text (EditorTextBufferSpellAdapte
                                                      guint                         offset,
                                                      guint                         length)
 {
-  _cjh_text_region_insert (self->region, offset, length, RUN_UNCHECKED);
+  if (self->enabled)
+    _cjh_text_region_insert (self->region, offset, length, RUN_UNCHECKED);
 }
 
 
@@ -675,7 +726,8 @@ editor_text_buffer_spell_adapter_before_delete_range (EditorTextBufferSpellAdapt
                                                       guint                         offset,
                                                       guint                         length)
 {
-  _cjh_text_region_remove (self->region, offset, length);
+  if (self->enabled)
+    _cjh_text_region_remove (self->region, offset, length);
 }
 
 void
@@ -687,17 +739,15 @@ editor_text_buffer_spell_adapter_after_delete_range (EditorTextBufferSpellAdapte
     mark_unchecked (self, offset, 0);
 }
 
-void
-editor_text_buffer_spell_adapter_cursor_moved (EditorTextBufferSpellAdapter *self,
-                                               guint                         position)
+static gboolean
+editor_text_buffer_spell_adapter_cursor_moved_cb (gpointer data)
 {
+  EditorTextBufferSpellAdapter *self = data;
   GtkTextIter begin, end;
 
-  g_return_if_fail (EDITOR_IS_TEXT_BUFFER_SPELL_ADAPTER (self));
-  g_return_if_fail (self->buffer != NULL);
+  g_assert (EDITOR_IS_TEXT_BUFFER_SPELL_ADAPTER (self));
 
-  if (self->cursor_position == position)
-    return;
+  self->queued_cursor_moved = 0;
 
   /* Invalidate the old position */
   if (self->enabled && get_word_at_position (self, self->cursor_position, &begin, &end))
@@ -705,12 +755,34 @@ editor_text_buffer_spell_adapter_cursor_moved (EditorTextBufferSpellAdapter *sel
                     gtk_text_iter_get_offset (&begin),
                     gtk_text_iter_get_offset (&end) - gtk_text_iter_get_offset (&begin));
 
-  self->cursor_position = position;
+  self->cursor_position = self->incoming_cursor_position;
 
-  if (self->enabled && get_word_at_position (self, position, &begin, &end))
+  /* Invalidate word at new position */
+  if (self->enabled && get_word_at_position (self, self->cursor_position, &begin, &end))
     mark_unchecked (self,
                     gtk_text_iter_get_offset (&begin),
                     gtk_text_iter_get_offset (&end) - gtk_text_iter_get_offset (&begin));
+
+  return G_SOURCE_REMOVE;
+}
+
+void
+editor_text_buffer_spell_adapter_cursor_moved (EditorTextBufferSpellAdapter *self,
+                                               guint                         position)
+{
+  g_return_if_fail (EDITOR_IS_TEXT_BUFFER_SPELL_ADAPTER (self));
+  g_return_if_fail (self->buffer != NULL);
+
+  if (!self->enabled)
+    return;
+
+  self->incoming_cursor_position = position;
+  g_clear_handle_id (&self->queued_cursor_moved, g_source_remove);
+  self->queued_cursor_moved = g_timeout_add_full (G_PRIORITY_LOW,
+                                                  INVALIDATE_DELAY_MSECS,
+                                                  editor_text_buffer_spell_adapter_cursor_moved_cb,
+                                                  g_object_ref (self),
+                                                  g_object_unref);
 }
 
 const char *
